// sound_meter.c
// for NerdKits with ATmega168
// mrobbins@mit.edu

#define F_CPU 14745600

#include <stdio.h>
#include <math.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>

#include "../libnerdkits/delay.h"
#include "../libnerdkits/uart.h"

#include "font.h"

// PIN DEFINITIONS:
//
// PC0 -- temperature sensor analog input

// PIN DEFINITIONS:
//
// PB1-5 ROW DRIVERS (0-4)
// PC1-5,PD2-7: COLUMN DRIVERS (1-11)
#define ROWS 5
#define COLS 22 //this is 2 less than the LED Marquee, we need one pin for the ADC

volatile uint8_t la_row, real_row;
volatile uint8_t la_data[COLS];

inline uint8_t ledarray_get(uint8_t i, uint8_t j) {
  if(i < ROWS && j < COLS) {
    if((la_data[j] & (1<<i)) != 0) {
      return 1;
    } else {
      return 0;
    }
  } else {
    return 0;
  }
}

inline void ledarray_set(uint8_t i, uint8_t j, uint8_t onoff) {
  if(i < ROWS && j < COLS) {
    if(onoff) {
      la_data[j] |= (1<<i);
    } else {
      la_data[j] &= ~(1<<i);
    }
  }
}

// col 0 - 21
inline void ledarray_set_column(uint8_t col, uint8_t onoff) {
  if ( col < COLS ) {
    if(onoff) {
      la_data[col] = 0xFF;
    } else {
      la_data[col] = 0x00;
    }

  }
}

inline void ledarray_set_columndriver(uint8_t j, uint8_t onoff, uint8_t sense) {
  // cols 0-4: PC1-5
  // cols 5-10: PD2-7
  if(j < 5) {
    if(onoff) {
      PORTC |= (1 << (PC1 + j));
    } else {
      PORTC &= ~(1<< (PC1 + j));
    }
    if(sense == onoff) {
      DDRC |= (1 << (PC1 + j));
    } else {
      DDRC &= ~(1 << (PC1 + j));
      PORTC &= ~(1 << (PC1 + j));
    }
  } else {
    if(onoff) {
      PORTD |= (1 << (PD2 + (j-5)));
    } else {
      PORTD &= ~(1<< (PD2 + (j-5)));
    }  
    if(sense == onoff) {
      DDRD |= (1 << (PD2 + (j-5)));
    } else {
      DDRD &= ~(1 << (PD2 + (j-5)));
      PORTD &= ~(1 << (PD2 + (j-5)));
    }
  }
}

inline void ledarray_all_off() {
  // turn off all row drivers
  DDRB &= ~( (1<<PB1)|(1<<PB2)|(1<<PB3)|(1<<PB4)|(1<<PB5) );
  PORTB &= ~( (1<<PB1)|(1<<PB2)|(1<<PB3)|(1<<PB4)|(1<<PB5) );
    
  // turn off all column drivers
  DDRC &= ~( (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5) );
  PORTC &= ~( (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5) );
  DDRD &= ~( (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) );
  PORTD &= ~( (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) );  
}

SIGNAL(SIG_OVERFLOW0) {
  // turn off old row driver
  DDRB &= ~(1 << (PB1 + real_row));
  PORTB &= ~(1 << (PB1 + real_row));
  ledarray_all_off();

  // increment row number
  if(++la_row == 2*ROWS)
    la_row = 0;

  // set column drivers appropriately
  uint8_t j;
  if(la_row%2 == 0) {
    // even la_row number: fill even columns
    real_row = la_row / 2;
    for(j=0; j<COLS/2; j++) {
      ledarray_set_columndriver(j, ledarray_get(real_row, 2*j), 1);
    }
    // activate row driver SINK
    PORTB &= ~(1 << (PB1 + real_row));
    DDRB |= (1 << (PB1 + real_row));
  } else {
    // odd la_row number: fill odd columns
    real_row = (la_row-1)/2;
    for(j=0; j<COLS/2; j++) {
      ledarray_set_columndriver(j, 1 - ledarray_get(real_row, 2*j + 1), 0);
    }
    // activate row driver SOURCE
    PORTB |= (1 << (PB1 + real_row));
    DDRB |= (1 << (PB1 + real_row));
  }  
}

void ledarray_init() {
  // Timer0 CK/64 (900Hz)
  TCCR0B = (1<<CS01) | (1<<CS00);
  TIMSK0 = (1<<TOIE0);
  
  // outputs (set row drivers high for off)
  DDRC &= ~( (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5) );
  DDRD &= ~( (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7) );
  DDRB &= ~( (1<<PB1)|(1<<PB2)|(1<<PB3)|(1<<PB4)|(1<<PB5) );
}

void ledarray_left_shift() {
  // shift everything one position left
  uint8_t i, j;
  for(i=0; i<ROWS; i++) {
    for(j=0; j<COLS-1; j++) {
      ledarray_set(i,j, ledarray_get(i, j+1));
    }
  }
}

void ledarray_blank() {
  uint8_t i, j;
  for(i=0; i<ROWS; i++) {
    for(j=0; j<COLS; j++) {
      ledarray_set(i,j,0);
    }
  }
}


void font_get(char match, char *buf) {
  // copies the character "match" into the buffer
  uint8_t i;
  PGM_P p;
  
  for(i=0; i<FONT_SIZE; i++) {
    memcpy_P(&p, &font[i], sizeof(PGM_P));
    
    if(memcmp_P(&match, p,1)==0) {
      memcpy_P(buf, p, 7);
      return;
    }
  }
  
  // NO MATCH?
  font_get('?', buf);
}

uint8_t font_width(char c) {
  char buf[7];
  buf[1] = 0;
  
  font_get(c, buf);
  
  return buf[1];
}

void font_display(char c, uint8_t offset) {
  char buf[7]; 
  font_get(c, buf);
  
  uint8_t width = buf[1];
  uint8_t i, j;
  for(i=0; i<ROWS; i++) {
    for(j=0; j<width; j++) {
      if((offset + j) < COLS) {
        if( (buf[2+j] & (1<<i)) != 0) {
          ledarray_set(i,offset + j,1);
        } else {
          ledarray_set(i,offset + j,0);
        }
      }
    }
  }
  
  // blank the next column to the right
  for(i=0; i<ROWS; i++) {
    ledarray_set(i, offset+width, 0);
  }
}

void do_scrolling_display() {
//  char nerdKits[12] = {'N','E','R','D','K','I','T','S','.','C','O','M'};
  int i;
  for(i=0;i<2;i++){ //scroll NERDKITS twice
    uint8_t strSize = 8;
    char nerdKits[strSize];
    nerdKits[0] = 'N';
    nerdKits[1] = 'E';
    nerdKits[2] = 'R';
    nerdKits[3] = 'D';
    nerdKits[4] = 'K';
    nerdKits[5] = 'I';
    nerdKits[6] = 'T';
    nerdKits[7] = 'S';

    ledarray_blank();
    uint8_t offset = 0;
    uint8_t next_offset = 0;
    char x=' ';
    
    uint8_t curr_char = 0;
    offset = next_offset = COLS;
    
    while(next_offset>0) {
      
      offset--;
      next_offset--;
      ledarray_left_shift();
      if(next_offset < COLS){
        if(curr_char<strSize){
          x = nerdKits[curr_char];
          curr_char++;
          offset = next_offset;
          next_offset = next_offset + font_width(x)+1;
        }
      }
      font_display(x,offset);
      delay_ms(200);
    }
  }
  
}

void adc_init() {
  // set analog to digital converter
  // for external reference (5v), single ended input ADC0
  ADMUX = 0;

  // set analog to digital converter
  // to be enabled, with a clock prescale of 1/16
  // so that the ADC clock runs at 8*115.2kHz.
  ADCSRA =  (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);

  // fire a conversion just to get the ADC warmed up
  ADCSRA |= (1<<ADSC);
}

uint16_t adc_read() {
  // read from ADC, waiting for conversion to finish
  // (assumes someone else asked for a conversion.)
  // wait for it to be cleared
  while(ADCSRA & (1<<ADSC)) {
    // do nothing... just hold your breath.
  }
  // bit is cleared, so we have a result.

  // read from the ADCL/ADCH registers, and combine the result
  // Note: ADCL must be read first (datasheet pp. 259)
  uint16_t result = ADCL;
  uint16_t temp = ADCH;
  result = result + (temp<<8);

  // set ADSC bit to get the *next* conversion started
  ADCSRA |= (1<<ADSC);
  
  return result;
}

void do_sound_meter(){
  uint16_t samples = 300;
  uint16_t max;
  uint16_t min;
  uint16_t i;
  //double old_mean=512, new_mean=512;
  //double rms;

  int8_t sticky_max = 0;
  uint8_t decay_rate = 1;
  int16_t this_sample;
  int16_t iterations = 1800;  
  for(;iterations > 0; iterations--){
    
    max = 0;
    min = 1024;
    for(i=0; i<samples; i++) { //take a set of samples, keep track of the min and max
      this_sample = adc_read();
      if(this_sample > max) {
        max = this_sample;
      } else if(this_sample < min) {
        min = this_sample;
      }
    }
    
    //diff between min and max represents sound level
    int16_t diff = max - min;

    uint8_t j;
    if(diff == 0) diff=1;
    double diff_d = 1.0* diff; //turn diff into a double
    double blocks_raw = 11.0*log10(diff_d) - 10.0; //take log of of the sound level and scale to fit onto the sound meter

    int8_t blocks;    
    if(blocks_raw >= 0.0)
      blocks = blocks_raw;
    else
      blocks = 0;

    //sticky max represents the extra bar at the top that slowly appears to get "pushed" up by the sound bar
    // set new sticky max
    if(blocks > sticky_max){
      sticky_max = blocks;
      decay_rate = 1;
    }

    //decay sticky max
    if(iterations % 4 == 0){
      sticky_max = sticky_max - decay_rate;
      decay_rate++; //increase the reate of falling so the bar appears to be accelarating down
    }

    //redner blocks to display
    for(j = COLS; j > (COLS-blocks); j--){
       ledarray_set_column(j-1,1);
    }
    for(; j > 0; j--){
       ledarray_set_column(j-1,0);
    }
    //render sticky max
    ledarray_set_column(COLS - sticky_max,1);

  } 
  return;
}

void do_bar_test(){
  uint8_t i = 0;
  while(1){
    for(i = 0; i < COLS; i++){
      ledarray_set_column(i,1);
      delay_ms(1000);
    }
    ledarray_blank();
    delay_ms(1000);
  }
}

int main() {
  ledarray_init();

  sei();

  // start up the Analog to Digital Converter
  adc_init();  
  
  while(1) {
      do_sound_meter();
      do_scrolling_display();
  }

}
